.code16
.section ".text"

#include "globals.h"

#define BLOCK_TO_SECTOR(x) (((x) * 4096) / 512 + 2048)
#define SECTOR_TO_BLOCK(x) (((x) - 2048) * 512 / 4096)

DISK_INT = 0x13 * 4
OLD_DISK_INT = 0x90 * 4
GRUB_MENU_LST_SECTOR = 985088

RAMDISK_IMAGE_OFFSET = 24
RAMDISK_SIZE_OFFSET = 28

BZIMAGE_FIRST_SECTOR = BLOCK_TO_SECTOR(113615)


.global _start
_start:
	mov $0, %ax
	mov %ax, %ds
	mov %ax, %es

	/* read second disk's MBR */
	mov $dap, %si
	movw $1, 2(%si)
	movw $0x7c00, 4(%si)
	movw $0, 6(%si)

	mov $0x42, %ah
	mov $0x81, %dl
	int $0x13

	mov $0, %ax
	mov %ax, %ds
	mov %ax, %es
	mov %ax, %ss

	/* overwrite int13h handler */
	movl DISK_INT, %eax
	movl %eax, OLD_DISK_INT

	movw $int13_handler, DISK_INT
	movw $STAGE2_SEG, DISK_INT + 2

	ljmp $0, $0x7c00

/*
 * check_dap_contains_sector(dap, n)
 * checks if dap->first_sector <= n <= dap->first_sector + dap->num_sectors
 *
 * sector	- 4(%bp)
 * dap pointer	- 8(%bp)
 */
check_dap_contains_sector:
	push %bp
	mov %sp, %bp

	push %si
	push %ecx
	push %edx

	/* sector */
	movl 4(%bp), %ecx
	/* dap pointer */
	movw 8(%bp), %si

	/* compare with low limit */
	movl 8(%si), %edx
	cmpl %ecx, %edx
	ja 1f

	/* compute high limit */
	xor %eax, %eax
	movw 2(%si), %ax
	addl %eax, %edx

	/* compare with high limit */
	cmpl %ecx, %edx
	jbe 1f

	pop %edx
	pop %ecx
	pop %si
	mov %bp, %sp
	pop %bp	
	mov $1, %eax
	ret $6

1:
	pop %edx
	pop %ecx	
	pop %si
	mov %bp, %sp
	pop %bp
	mov $0, %eax
	ret $6

/*
 * update_data(dap, sector, offset, data, len)
 *
 * memcpy(dap->buffer + (sector - dap->start_sector) * 512 + offset, data, len)
 *
 * dap pointer - 4(%bp)
 * sector - 6(%bp)
 * offset - 10(%bp)
 * data_off - 14(%bp)
 * data_seg - 16(%bp)	
 * len - 18(%bp)
 */
update_dap_data:
	push %bp
	mov %sp, %bp

	push %ebx
	push %ecx
	push %edi
	push %si
	push %ds
	push %es

	xor %eax, %eax
	xor %edx, %edx
	
	/* dap */
	mov 4(%bp), %di
	/* sector */
	mov 6(%bp), %ebx

	/*
	 * compute offset in bytes
	 * offset = (sector - dap->start_sector) << 9
	 */
	sub 8(%di), %ebx
	shl $9, %ebx
	addl 10(%bp), %ebx

	/* normalize dest pointer */
	mov 6(%di), %ax
	shl $4, %eax
	mov 4(%di), %dx
	add %edx, %eax

	/* add offset */
	add %ebx, %eax

	/* convert back to seg:offset */
	mov %eax, %edi
	and $0xf, %edi
	shr $4, %eax
	mov %ax, %es

	/* ds:si come as parameters */
	pushw 16(%bp)
	pop %ds

	mov 14(%bp), %si

	/* len also as parameter */
	mov 18(%bp), %ecx

	rep movsb

	pop %es
	pop %ds
	pop %si
	pop %edi
	pop %ecx
	pop %ebx

	mov %bp, %sp
	pop %bp

	ret $18

/* ---------------------------------------------------
 * int 13h handler starts here
 */
int13_handler:
	/* not a read, go to old handler */
	cmp $0x42, %ah
	jne do_old_handler

	pushf
	pushal

/* reading a range which includes the grub menu.lst sector */	
check_menu_lst:
	push %eax
	push %edx

	pushw %si
	pushl $GRUB_MENU_LST_SECTOR
	call check_dap_contains_sector
	test %eax, %eax
	jz check_bzImage_header

	pop %edx
	pop %eax

	/* call old handler to read the data */
	int $0x90

	pushl $menu_lst_len		/* len */
	pushw $STAGE2_SEG			/* buffer seg */
	pushw $menu_lst			/* buffer offset */
	pushl $0			/* menu.lst starts at sector boundary */
	pushl $GRUB_MENU_LST_SECTOR	/* menu.lst sector */
	push %si
	call update_dap_data

	popal
	popf

	clc
	mov $0, %ah

	lret $0x2

check_bzImage_header:
	push %si
	pushl $(1 + BZIMAGE_FIRST_SECTOR) /* bzImage ramdisk info is in the 2nd sector */
	call check_dap_contains_sector
	test %eax, %eax
	jz done

	pop %edx
	pop %eax

	int $0x90

	/* eax = (dap->sector - (1 + bzimage_first_sector)) * 512 + dap->offset + RAMDISK_IMAGE_OFFSET */
	mov $(1 + BZIMAGE_FIRST_SECTOR), %eax
	sub 8(%si), %eax
	shl $9, %eax

	add 4(%si), %ax
	add $RAMDISK_IMAGE_OFFSET, %eax	

	push %es
	mov 6(%si), %es

	/* ebx - ramdisk address */
	mov $STAGE2_SEG, %ebx
	shl $4, %ebx
	add $(STAGE2_ADDR + STAGE2_NUM_SECTORS * 512), %ebx

	movl %ebx, %es:(%eax)

	sub $RAMDISK_IMAGE_OFFSET, %eax
	add $RAMDISK_SIZE_OFFSET, %eax
	movl $INITRD_SIZE, %es:(%eax)

	pop %es

	popal
	popf

	clc
	mov $0, %ah

	lret $0x2
	
done:
	pop %edx
	pop %eax
	
	popal
	popf

do_old_handler:
	int $0x90
	lret $0x2


dap:
	.byte 0x10	/* size */
	.byte 0		/* sbz */
	.word 0		/* num sectors */
	.word 0		/* buffer offset */
	.word 0		/* buffer segment */
	.long 0		/* first sector low */
	.long 0		/* first sector high */

menu_lst:
	.asciz "timeout 3\ntitle IO resident challenge\nroot (hd1,0)\nkernel /vmlinuz root=/dev/sdb1\n"
menu_lst_len = . - menu_lst

.align 512
STAGE2_NUM_SECTORS = (. - _start) / 512
initrd_start:
.incbin "initrd/initrd.cpio.gz"
initrd_end:
INITRD_SIZE = initrd_end - initrd_start
.align 512

. = _start + 2047*512
.incbin	"pic.bin"
